"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.GCDataBuilder = exports.concatGarbageCollectionData = exports.concatGarbageCollectionStates = exports.removeRouteFromAllNodes = exports.unpackChildNodesUsedRoutes = exports.unpackChildNodesGCDetails = exports.cloneGCData = exports.trimTrailingSlashes = exports.trimLeadingSlashes = exports.trimLeadingAndTrailingSlashes = void 0;
const common_utils_1 = require("@fluidframework/common-utils");
/**
 * Trims the leading and trailing slashes from the given string.
 * @param str - A string that may contain leading and / or trailing slashes.
 * @returns A new string without leading and trailing slashes.
 */
function trimLeadingAndTrailingSlashes(str) {
    return str.replace(/^\/+|\/+$/g, "");
}
exports.trimLeadingAndTrailingSlashes = trimLeadingAndTrailingSlashes;
/**
 * Trims the leading slashes from the given string.
 * @param str - A string that may contain leading slashes.
 * @returns A new string without leading slashes.
 */
function trimLeadingSlashes(str) {
    return str.replace(/^\/+/g, "");
}
exports.trimLeadingSlashes = trimLeadingSlashes;
/**
 * Trims the trailing slashes from the given string.
 * @param str - A string that may contain trailing slashes.
 * @returns A new string without trailing slashes.
 */
function trimTrailingSlashes(str) {
    return str.replace(/\/+$/g, "");
}
exports.trimTrailingSlashes = trimTrailingSlashes;
/**
 * Helper function that clones the GC data.
 * @param gcData - The GC data to clone.
 * @returns a clone of the given GC data.
 */
function cloneGCData(gcData) {
    const clonedGCNodes = {};
    for (const [id, outboundRoutes] of Object.entries(gcData.gcNodes)) {
        clonedGCNodes[id] = Array.from(outboundRoutes);
    }
    return {
        gcNodes: clonedGCNodes,
    };
}
exports.cloneGCData = cloneGCData;
/**
 * Helper function that unpacks the GC details of the children from a given node's GC details.
 * @param gcDetails - The GC details of a node.
 * @returns A map of GC details of each children of the the given node.
 */
function unpackChildNodesGCDetails(gcDetails) {
    const childGCDetailsMap = new Map();
    // If GC data is not available, bail out.
    if (gcDetails.gcData === undefined) {
        return childGCDetailsMap;
    }
    // Remove the node's self GC nodes, if any, and generate the children GC nodes.
    const gcNodes = gcDetails.gcData.gcNodes;
    delete gcNodes["/"];
    for (const [id, outboundRoutes] of Object.entries(gcNodes)) {
        (0, common_utils_1.assert)(id.startsWith("/"), 0x2ae /* "node id should always be an absolute route" */);
        const childId = id.split("/")[1];
        let childGCNodeId = id.slice(childId.length + 1);
        // GC node id always begins with "/". Handle the special case where a child's id in the parent's GC nodes is
        // of format `/root`. In this case, the childId is root and childGCNodeId is "". Make childGCNodeId = "/".
        if (childGCNodeId === "") {
            childGCNodeId = "/";
        }
        let childGCDetails = childGCDetailsMap.get(childId);
        if (childGCDetails === undefined) {
            childGCDetails = { gcData: { gcNodes: {} }, usedRoutes: [] };
        }
        // gcData should not undefined as its always at least initialized as  empty above.
        (0, common_utils_1.assert)(childGCDetails.gcData !== undefined, 0x2af /* "Child GC data should have been initialized" */);
        childGCDetails.gcData.gcNodes[childGCNodeId] = [...new Set(outboundRoutes)];
        childGCDetailsMap.set(childId, childGCDetails);
    }
    if (gcDetails.usedRoutes === undefined) {
        return childGCDetailsMap;
    }
    // Remove the node's self used route, if any, and generate the children used routes.
    const usedRoutes = gcDetails.usedRoutes.filter((route) => route !== "" && route !== "/");
    for (const route of usedRoutes) {
        (0, common_utils_1.assert)(route.startsWith("/"), 0x2b0 /* "Used route should always be an absolute route" */);
        const childId = route.split("/")[1];
        const childUsedRoute = route.slice(childId.length + 1);
        const childGCDetails = childGCDetailsMap.get(childId);
        (0, common_utils_1.assert)((childGCDetails === null || childGCDetails === void 0 ? void 0 : childGCDetails.usedRoutes) !== undefined, 0x2b1 /* "This should have be initiallized when generate GC nodes above" */);
        childGCDetails.usedRoutes.push(childUsedRoute);
        childGCDetailsMap.set(childId, childGCDetails);
    }
    return childGCDetailsMap;
}
exports.unpackChildNodesGCDetails = unpackChildNodesGCDetails;
/**
 * Helper function that unpacks the used routes of children from a given node's used routes.
 * @param usedRoutes - The used routes of a node.
 * @returns A map of used routes of each children of the the given node.
 */
function unpackChildNodesUsedRoutes(usedRoutes) {
    // Remove the node's self used route, if any, and generate the children used routes.
    const filteredUsedRoutes = usedRoutes.filter((route) => route !== "" && route !== "/");
    const childUsedRoutesMap = new Map();
    for (const route of filteredUsedRoutes) {
        (0, common_utils_1.assert)(route.startsWith("/"), 0x198 /* "Used route should always be an absolute route" */);
        const childId = route.split("/")[1];
        const childUsedRoute = route.slice(childId.length + 1);
        const childUsedRoutes = childUsedRoutesMap.get(childId);
        if (childUsedRoutes !== undefined) {
            childUsedRoutes.push(childUsedRoute);
        }
        else {
            childUsedRoutesMap.set(childId, [childUsedRoute]);
        }
    }
    return childUsedRoutesMap;
}
exports.unpackChildNodesUsedRoutes = unpackChildNodesUsedRoutes;
/**
 * Removes the given route from the outbound routes of all the given GC nodes, and any duplicates
 * @param gcNodes - The nodes from which the route is to be removed.
 * @param outboundRoute - The route to be removed.
 */
function removeRouteFromAllNodes(gcNodes, outboundRoute) {
    const channels = Object.entries(gcNodes);
    for (const [nodeId, outboundRoutes] of channels) {
        // Remove route from channel to parent for each channel
        const outboundRoutesSet = new Set(outboundRoutes);
        outboundRoutesSet.delete(outboundRoute);
        gcNodes[nodeId] = [...outboundRoutesSet];
    }
}
exports.removeRouteFromAllNodes = removeRouteFromAllNodes;
/**
 * Concatenates the given GC states and returns the concatenated GC state.
 */
function concatGarbageCollectionStates(gcState1, gcState2) {
    var _a;
    const combinedGCNodes = {};
    for (const [nodeId, nodeData] of Object.entries(gcState1.gcNodes)) {
        combinedGCNodes[nodeId] = {
            outboundRoutes: Array.from(nodeData.outboundRoutes),
            unreferencedTimestampMs: nodeData.unreferencedTimestampMs,
        };
    }
    for (const [nodeId, nodeData] of Object.entries(gcState2.gcNodes)) {
        let combinedNodedata = combinedGCNodes[nodeId];
        if (combinedNodedata === undefined) {
            combinedNodedata = {
                outboundRoutes: Array.from(nodeData.outboundRoutes),
                unreferencedTimestampMs: nodeData.unreferencedTimestampMs,
            };
        }
        else {
            // Validate that same node doesn't have different unreferenced timestamp.
            if (nodeData.unreferencedTimestampMs !== undefined
                && combinedNodedata.unreferencedTimestampMs !== undefined) {
                (0, common_utils_1.assert)(nodeData.unreferencedTimestampMs === combinedNodedata.unreferencedTimestampMs, 0x2b2 /* "Two entries for the same GC node with different unreferenced timestamp" */);
            }
            combinedNodedata = {
                outboundRoutes: [...new Set([...nodeData.outboundRoutes, ...combinedNodedata.outboundRoutes])],
                unreferencedTimestampMs: (_a = nodeData.unreferencedTimestampMs) !== null && _a !== void 0 ? _a : combinedNodedata.unreferencedTimestampMs,
            };
        }
        combinedGCNodes[nodeId] = combinedNodedata;
    }
    return { gcNodes: combinedGCNodes };
}
exports.concatGarbageCollectionStates = concatGarbageCollectionStates;
/**
 * Concatenates the given GC datas and returns the concatenated GC data.
 */
function concatGarbageCollectionData(gcData1, gcData2) {
    const combinedGCData = cloneGCData(gcData1);
    for (const [id, routes] of Object.entries(gcData2.gcNodes)) {
        if (combinedGCData.gcNodes[id] === undefined) {
            combinedGCData.gcNodes[id] = Array.from(routes);
        }
        else {
            const combinedRoutes = [...routes, ...combinedGCData.gcNodes[id]];
            combinedGCData.gcNodes[id] = [...new Set(combinedRoutes)];
        }
    }
    return combinedGCData;
}
exports.concatGarbageCollectionData = concatGarbageCollectionData;
class GCDataBuilder {
    constructor() {
        this.gcNodesSet = {};
    }
    get gcNodes() {
        const gcNodes = {};
        for (const [nodeId, outboundRoutes] of Object.entries(this.gcNodesSet)) {
            gcNodes[nodeId] = [...outboundRoutes];
        }
        return gcNodes;
    }
    addNode(id, outboundRoutes) {
        this.gcNodesSet[id] = new Set(outboundRoutes);
    }
    /**
     * Adds the given GC nodes. It does the following:
     * - Normalizes the ids of the given nodes.
     * - Prefixes the given `prefixId` to the given nodes' ids.
     * - Adds the outbound routes of the nodes against the normalized and prefixed id.
     */
    prefixAndAddNodes(prefixId, gcNodes) {
        for (const [id, outboundRoutes] of Object.entries(gcNodes)) {
            // Remove any leading slashes from the id.
            let normalizedId = trimLeadingSlashes(id);
            // Prefix the given id to the normalized id.
            normalizedId = `/${prefixId}/${normalizedId}`;
            // Remove any trailing slashes from the normalized id. Note that the trailing slashes are removed after
            // adding the prefix for handling the special case where id is "/".
            normalizedId = trimTrailingSlashes(normalizedId);
            // Add the outbound routes against the normalized and prefixed id without duplicates.
            this.gcNodesSet[normalizedId] = new Set(outboundRoutes);
        }
    }
    addNodes(gcNodes) {
        for (const [id, outboundRoutes] of Object.entries(gcNodes)) {
            this.gcNodesSet[id] = new Set(outboundRoutes);
        }
    }
    /**
     * Adds the given outbound route to the outbound routes of all GC nodes.
     */
    addRouteToAllNodes(outboundRoute) {
        for (const outboundRoutes of Object.values(this.gcNodesSet)) {
            outboundRoutes.add(outboundRoute);
        }
    }
    getGCData() {
        return {
            gcNodes: this.gcNodes,
        };
    }
}
exports.GCDataBuilder = GCDataBuilder;
//# sourceMappingURL=utils.js.map